include "programming_language.dfy"
include "util.dfy"

module assertions {
    import opened programming_language
    import opened util

    // ================================================================
    // Syntax of the assertion language
    // ================================================================

    datatype Assertion = Bool(BExpr) | emp_h | emp_s | Own(VName)
        | Points(Expr, Expr) | Reserved(Expr)
        | Indep(Assertion, Assertion) | Wand(Assertion, Assertion)
        | And(Assertion, Assertion) | Or(Assertion, Assertion)
        | Neg(Assertion) | Exists(VName, Assertion)
        | Impl(Assertion, Assertion)

    predicate SemanticAssertion(assrt: Assertion, mem: Mem)
    {
        var s := mem.0;
        var i := mem.1;
        var h := mem.2;
        match assrt
        case Bool(b) => DefinedBExpr(b, s, i) && ValBExpr(b, s, i)
        case emp_h => h == map[]
        case emp_s => s == map[]
        case Own(x) => domain(s) == {x}
        case Points(e1, e2) =>
            DefinedExpr(e1, s, i)
            && DefinedExpr(e2, s, i)
            && domain(h) == {ValExpr(e1, s, i)}
            && h[ValExpr(e1, s, i)] == ValExpr(e2, s, i)
        case Reserved(e) =>
            DefinedExpr(e, s, i)
            && domain(h) == {ValExpr(e, s, i)}
        case Indep(p, q) =>
            exists m, m' :: MemUnionable(m, m')
            && mem == MemUnion(m, m')
            && SemanticAssertion(p, m)
            && SemanticAssertion(q, m')
        case Wand(p, q) =>
            exists m, m' :: MemUnionable(mem, m)
            && m' == MemUnion(m, mem)
            && SemanticAssertion(p, m)
            && SemanticAssertion(q, m')
        case And(p, q) => SemanticAssertion(p, mem) && SemanticAssertion(q, mem)
        case Or(p, q) => SemanticAssertion(p, mem) || SemanticAssertion(q, mem)
        case Neg(p) => ! SemanticAssertion(p, mem)
        case Exists(X, p) =>
            exists v :: SemanticAssertion(p, (s, i[X := v], h))
        case Impl(p, q) =>
            SemanticAssertion(p, mem) ==> SemanticAssertion(q, mem)
    }

    function OwnFromSeq(sn: seq<VName>): Assertion
    {
        if |sn| == 0 then emp_s()
        else if |sn| == 1 then Own(sn[0])
        else Assertion.Indep(Own(sn[0]), OwnFromSeq(sn[1..]))
    }

    function AndFromSeq(sa: seq<Assertion>): Assertion
        requires |sa| > 0
    {
        if |sa| == 1 then sa[0]
        else Assertion.And(sa[0], AndFromSeq(sa[1..]))
    }

    function OrFromSeq(sa: seq<Assertion>): Assertion
        requires |sa| > 0
    {
        if |sa| == 1 then sa[0]
        else Assertion.Or(sa[0], OrFromSeq(sa[1..]))
    }

    function IndepFromSeq(sa: seq<Assertion>): Assertion
        requires |sa| > 0
    {
        if |sa| == 1 then sa[0]
        else Assertion.Indep(sa[0], IndepFromSeq(sa[1..]))
    }


    function MultiplePoints(e: Expr, se: seq<Expr>): Assertion
        requires |se| > 0
    {
        helper_MultiplePoints(e, se, 0)
    }

    function helper_MultiplePoints(e: Expr, se: seq<Expr>, n: int): Assertion
        requires |se| > 0
    {
        if |se| == 1 then Points(Add(e, Int(n)), se[0])
        else Assertion.And(Points(Add(e, Int(n)), se[0]),
            helper_MultiplePoints(e, se[1..], n+1))
    }

    predicate AssertionImpl(p: Assertion, q: Assertion) {
        forall m :: SemanticAssertion(Assertion.Impl(p, q), m)
    }

    function FvAssertion(assrt: Assertion):set<VName> {
        match assrt
        case Bool(b) => FvBExpr(b)
        case emp_h => {}
        case emp_s => {}
        case Own(x) => {x}
        case Points(e1, e2) => FvExpr(e1) + FvExpr(e2)
        case Reserved(e) => FvExpr(e)
        case Indep(p, q) => FvAssertion(p) + FvAssertion(q)
        case Wand(p, q) => FvAssertion(p) + FvAssertion(q)
        case And(p, q) => FvAssertion(p) + FvAssertion(q)
        case Or(p, q) => FvAssertion(p) + FvAssertion(q)
        case Neg(p) => FvAssertion(p)
        case Exists(X, p) => FvAssertion(p)
        case Impl(p, q) => FvAssertion(p) + FvAssertion(q)
    }



    function emp(): Assertion {
        Assertion.And(emp_s, emp_h)
    }

    datatype Action = Change(Assertion, Assertion) | Mantain(Assertion)
        | Indep(Action, Action) | Exists(VName, Action)
        | Impl(Action, Action) | And(Action, Action) | Or(Action, Action)

    predicate SemanticAction(assrt: Action, m: Mem, m': Mem)
    {
        var s := m.0;
        var i := m.1;
        var h := m.2;
        var s' := m'.0;
        var i' := m'.1;
        var h' := m'.2;

        match assrt
        case Change(p, q) => SemanticAssertion(p, m) && SemanticAssertion(q, m')
        case Mantain(p) => m == m' && SemanticAssertion(p, m)
        case Indep(a1, a2) =>
            exists m1, m2, m1', m2' ::
                MemUnionable(m1, m2)
                && m == MemUnion(m1, m2)
                && MemUnionable(m1', m2')
                && m' == MemUnion(m1', m2')
                && SemanticAction(a1, m1, m1')
                && SemanticAction(a2, m2, m2')
        case Exists(X, a) =>
            i == i'
            && exists v :: SemanticAction(a, (s, i[X := v], h), (s', i[X := v], h'))
        case And(a1, a2) => SemanticAction(a1, m, m') && SemanticAction(a2, m, m')
        case Or(a1, a2) => SemanticAction(a1, m, m') || SemanticAction(a2, m, m')
        case Impl(a1, a2) => SemanticAction(a1, m, m') ==> SemanticAction(a2, m, m')
    }

    predicate ActionImpl(a1: Action, a2: Action) {
        forall m, m' :: SemanticAction(Action.Impl(a1, a2), m, m')
    }

    function FvAction(act: Action):set<VName> {
        match act
        case Change(p, q) => FvAssertion(p) + FvAssertion(q)
        case Mantain(p) => FvAssertion(p)
        case Indep(a1, a2) => FvAction(a1) + FvAction(a2)
        case Exists(X, a) => FvAction(a)
        case And(a1, a2) => FvAction(a1) + FvAction(a2)
        case Or(a1, a2) => FvAction(a1) + FvAction(a2)
        case Impl(a1, a2) => FvAction(a1) + FvAction(a2)
    }

    function Id(): Action {
        Mantain(Bool(True))
    }

    function TrueAct(): Action {
        Change(Bool(True), Bool(True))
    }

    function Emp(): Action {
        Change(emp(), emp())
    }


    function MemUnion(m: Mem, m': Mem): Mem
        requires MemUnionable(m, m')
    {
        var s := m.0;
        var i := m.1;
        var h := m.2;
        var s' := m'.0;
        var i' := m'.1;
        var h' := m'.2;
        (union(s, s'), i, union(h, h'))
    }

    predicate MemUnionable(m: Mem, m': Mem) {
        var s := m.0;
        var i := m.1;
        var h := m.2;
        var s' := m'.0;
        var i' := m'.1;
        var h' := m'.2;
        i == i' && s !! s' && h !! h'
    }
}

