include "programming_language.dfy"
include "assertions.dfy"
include "util.dfy"

module inference_rules {
    import opened programming_language
    import opened assertions
    import opened util

    predicate Sta(p: Assertion, a: Action)
    {
        forall m, m' ::
            SemanticAssertion(p, m)
            && SemanticAction(a, m, m')
            ==>
            SemanticAssertion(p, m')
    }

    lemma Sta_Equivalence1(r: Assertion, p: Assertion, q: Assertion)
        ensures
            Sta(r, Change(p, q))
            ==>
            (forall m ::
                SemanticAssertion(Assertion.Indep(
                    Assertion.And(Wand(p, r), emp()),
                    q), m)
                ==>
                SemanticAssertion(r, m))

    lemma Sta_Equivalence2(r: Assertion, p: Assertion, q: Assertion)
        ensures
            Sta(r, Action.Indep(Change(p, q), Id()))
            ==>
            (forall m ::
                SemanticAssertion(Assertion.Indep(Wand(p, r), q), m)
                ==>
                SemanticAssertion(r, m))

    lemma Sta_Implicate(r: Assertion, p: Assertion, q: Assertion)
        ensures
            (forall m ::
                SemanticAssertion(Assertion.Indep(Wand(p, r), q), m)
                ==>
                SemanticAssertion(r, m))
            ==>
            Sta(r, Change(p, q))

    predicate Precise(p: Assertion) {
        forall s: Store, i, h: Heap, s1, s2, h1, h2 ::
            domain(s1) <= domain(s)
            && domain(s2) <= domain(s)
            && domain(h1) <= domain(h)
            && domain(h2) <= domain(h)
            && SemanticAssertion(p, (s1, i, h1))
            && SemanticAssertion(p, (s2, i, h2))
            ==>
            s1 == s2
            && h1 == h2
    }


    predicate Fence(I: Assertion, a: Action) {
        forall m, m' ::
            SemanticAction(Action.Impl(Mantain(I), a), m, m')
            && SemanticAction(Action.Impl(a, Change(I, I)), m, m')
            && Precise(I)
    }

    lemma Sta_Separation(p: Assertion, a: Action,
            p': Assertion, a': Action, I: Assertion)
        requires Sta(p, a) && Sta(p', a')
            && AssertionImpl(p, I)
            && Fence(I, a)
        ensures Sta(Assertion.Indep(p, p'), Action.Indep(a, a'))



    // ========================================================================
    // Sequential inference rules
    // ========================================================================

    inductive predicate SequentiallyDerivable(p: Assertion, C: Stmt, q: Assertion) {
        (match C
        // ASSN
        case Assign(x, E) => (exists X, owned ::
            x in owned
            && p == Assertion.And(OwnFromSeq(owned), Assertion.And(Bool(Equal(LVar(X), E)), emp_h))
            && q == Assertion.And(OwnFromSeq(owned), Assertion.And(Bool(Equal(Var(x), LVar(X))), emp_h))
        )
        // LD
        case Load(x, E) => (exists X, Y, owned ::
            x in owned
            && p == Assertion.And(OwnFromSeq(owned),
                Assertion.And(Bool(Equal(Var(x), LVar(X))), Points(E, Y)))
            && q == Assertion.And(OwnFromSeq(owned),
                Assertion.And(Bool(Equal(Var(x), Y)), Points(ReplaceVar(E, x, X), Y)))
        )
        // ST
        case Store(E1, E2) => (exists owned ::
            (AssertionImpl(Reserved(E1), Bool(Equal(E2, E2))))
            && p == Assertion.And(OwnFromSeq(owned), Reserved(E1))
            && q == Assertion.And(OwnFromSeq(owned), Points(E1, E2))
        )
        // DISP
        case Dispose(E) => (exists owned ::
            q == Assertion.And(OwnFromSeq(owned), Reserved(E))
            && q == Assertion.And(OwnFromSeq(owned), emp_h)
        )
        // CONS
        case Cons(x, se) => (
            (forall m: Mem ::
                (exists X ::
                    SemanticAssertion(
                        Assertion.And(Own(x), Assertion.And(Bool(Equal(Var(x), LVar(X))), emp_h)), m)
                    ==>
                    helper_AllDefined(se, m.0, m.1)
                )
            )
            && (exists X, Y, owned ::
                x in owned
                && p == Assertion.And(OwnFromSeq(owned),
                    Assertion.And(Bool(Equal(Var(x), LVar(X))), emp_h))
                && q == Assertion.And(OwnFromSeq(owned),
                    Assertion.And(Bool(Equal(Var(x), LVar(Y))),
                    helper_ConsPost(se, x, X, Y, 0)))
            )
        )
        // SKIP
        case Skip => (
            p == q
        )
        // SEQ
        case Seq(C1, C2) => (
            exists p' :: SequentiallyDerivable(p, C1, p')
                && SequentiallyDerivable(p', C2, q)
        )
        // IF
        case If(B, C1, C2) => (
            (forall m :: SemanticAssertion(p, m) ==> DefinedBExpr(B, m.0, m.1))
            && SequentiallyDerivable(Assertion.And(p, Bool(B)), C1, q)
            && SequentiallyDerivable(Assertion.And(p, Bool(Not(B))), C2, q)
        )
        // WHILE
        case While(B, C) => (
            (forall m :: SemanticAssertion(p, m) ==> DefinedBExpr(B, m.0, m.1))
            && SequentiallyDerivable(Assertion.And(p, Bool(B)), C, p)
            &&  q == Assertion.And(p, Bool(Not(B)))
        )
        // Not used ====================
        case Atomic(_, _) => false
        case Par(_, _) => false
        // End not used ================
        )
        // CONSEQ
        || (
            exists p', q' ::
                AssertionImpl(p, p')
                && AssertionImpl(q', q)
                && SequentiallyDerivable(p', C, q')
        )
        // FRM
        || (
            exists r, p', q' ::
                p == Assertion.Indep(p', r)
                && q == Assertion.Indep(q', r)
                && SequentiallyDerivable(p', C, q')
        )
        // CONJ
        || (
            exists p1, p2, q1, q2 ::
                p == Assertion.And(p1, p2)
                && q == Assertion.And(q1, q2)
                && SequentiallyDerivable(p1, C, q1)
                && SequentiallyDerivable(p2, C, q2)
        )
        // DISJ
        || (
            exists p1, p2, q1, q2 ::
                p == Assertion.Or(p1, p2)
                && q == Assertion.Or(q1, q2)
                && SequentiallyDerivable(p1, C, q1)
                && SequentiallyDerivable(p2, C, q2)
        )
        // EXISTS
        || (
            exists p', q', X ::
                p == Assertion.Exists(X, p')
                && q == Assertion.Exists(X, q')
                && SequentiallyDerivable(p', C, q')
        )
    }

    predicate helper_AllDefined(se: seq<Expr>, s: Store, i: LvMap) {
        if |se| == 0 then true
        else DefinedExpr(se[0], s, i) && helper_AllDefined(se[1..], s, i)
    }

    function helper_ConsPost(se: seq<Expr>, x: VName, X: VName, Y: VName, ind: int): Assertion {
        if |se| == 0 then Bool(True)
        else Assertion.And(
            Points(Add(LVar(Y), Int(ind)), ReplaceVar(se[0], x, X)),
            helper_ConsPost(se[1..], x, X, Y, ind+1)
        )
    }


    // ========================================================================
    // Inference rules for concurrency
    // ========================================================================

    inductive predicate ConcurrentlyDerivable(R: Action, G: Action, I: Assertion,
            p: Assertion, C: Stmt, q: Assertion)
    {
        (match C
        // P-SEQ
        case Seq(C1, C2) => (
            exists p' :: ConcurrentlyDerivable(R, G, I, p, C1, p')
                && ConcurrentlyDerivable(R, G, I, p', C2, q)
        )
        // ATOMIC
        case Atomic(B, C) => (
            (forall m :: SemanticAssertion(p, m) ==> DefinedBExpr(B, m.0, m.1))
            && SequentiallyDerivable(Assertion.And(p, Bool(B)), C, q)
            && Sta(p, Action.Indep(R, Id()))
            && Sta(q, Action.Indep(R, Id()))
            && ActionImpl(Change(p, q), Action.Indep(G, TrueAct()))
            && AssertionImpl(Assertion.Or(p, q), Assertion.Indep(I, Bool(True)))
            && Fence(I, R)
            && Fence(I, G)
        )
        // P-WHILE
        case While(B, C) => (
            // P ==> (B = B)*I
            AssertionImpl(p, Assertion.Indep(Assertion.Or(Bool(B), Bool(Not(B))), I))
            && ConcurrentlyDerivable(R, G, I, Assertion.And(p, Bool(B)), C, p)
            && q == Assertion.And(p, Bool(Not(B)))
        )
        // P-IF
        case If(B, C1, C2) => (
            // P ==> (B = B)*I
            AssertionImpl(p, Assertion.Indep(Assertion.Or(Bool(B), Bool(Not(B))), I))
            && ConcurrentlyDerivable(R, G, I, Assertion.And(p, Bool(B)), C1, q)
            && ConcurrentlyDerivable(R, G, I, Assertion.And(p, Bool(Not(B))), C2, q)
        )
        // PAR
        case Par(C1, C2) => (exists p1, p2, q1, q2, r, r1, r2, G1, G2 ::
            G == Action.Or(G1, G2)
            && p == Assertion.Indep(p1, Assertion.Indep(p2, r))
            && q == Assertion.Indep(p1, Assertion.Indep(p2, Assertion.And(r1, r2)))
            && ConcurrentlyDerivable(
                    Action.Or(R, G2), G1, I,
                    Assertion.Indep(p1, r), C, Assertion.Indep(q1, r1)
            )
            && ConcurrentlyDerivable(
                    Action.Or(R, G1), G2, I,
                    Assertion.Indep(p2, r), C, Assertion.Indep(q2, r2)
            )
            && AssertionImpl(Assertion.Or(r, Assertion.Or(r1, r2)), I)
            && Fence(I, R)
        )
        // Not used ====================
        case Assign(_, _) => false
        case Load(_, _) => false
        case Store(_, _) => false
        case Skip => false
        case Cons(_, _) => false
        case Dispose(_) => false
        // End not used ================
        )
        // ENV
        || (
            R == Emp() && G == Emp() && I == emp()
            && SequentiallyDerivable(p, C, q)
        )
        // FRAME
        || (exists R', R'', G', G'', I', I'', p', q', r ::
            R == Action.Indep(R', R'')
            && G == Action.Indep(G', G'')
            && I == Assertion.Indep(I', I'')
            && p == Assertion.Indep(p', r)
            && q == Assertion.Indep(q', r)

            && ConcurrentlyDerivable(R', G', I', p', C, q')
            && Sta(r, Action.Indep(R'', Id()))
            && Fence(I'', R'')
            && Fence(I'', G'')
            && AssertionImpl(r, Assertion.Indep(I'', Bool(True)))
        )
        // HIDE
        || (exists R', G', I' ::
            ConcurrentlyDerivable(
                Action.Indep(R, R'), Action.Indep(G, G'),
                Assertion.Indep(I, I'), p, C, q)
            && Fence(I, R)
            && Fence(I, G)
        )
        // P-EX
        || (exists p', q', X ::
            p == Assertion.Exists(X, p')
            && q == Assertion.Exists(X, q')
            && ConcurrentlyDerivable(R, G, I, p', C, q')
            && X !in FvAction(R)
            && X !in FvAction(G)
            && X !in FvAssertion(I)
        )
        // P-CONJ
        || (exists p1, p2, q1, q2 ::
            p == Assertion.And(p1, p2)
            && q == Assertion.And(q1, q2)
            && ConcurrentlyDerivable(R, G, I, p1, C, q1)
            && ConcurrentlyDerivable(R, G, I, p2, C, q2)
        )
        // DISJ
        || (exists p1, p2, q1, q2 ::
            p == Assertion.Or(p1, p2)
            && q == Assertion.Or(q1, q2)
            && ConcurrentlyDerivable(R, G, I, p1, C, q1)
            && ConcurrentlyDerivable(R, G, I, p2, C, q2)
        )
        // CSQ
        || (exists p', R', G', q', I' ::
            AssertionImpl(p, p')
            && ActionImpl(R, R')
            && ActionImpl(G', G)
            && AssertionImpl(q', q)
            && ConcurrentlyDerivable(R', G', I', p', C, q')
            && AssertionImpl(Assertion.Or(p, q), Assertion.Indep(I, Bool(True)))
            && Fence(I, R)
            && Fence(I, G)
        )
        // Allow derived rules
        || (
            ConcurrentlyDerivable_derived(R, G, I, p, C, q)
        )
    }

    inductive predicate ConcurrentlyDerivable_derived(R: Action, G: Action, I: Assertion,
            p: Assertion, C: Stmt, q: Assertion) {

        (match C
        // Not used ====================
        case Assign(_, _) => false
        case Load(_, _) => false
        case Store(_, _)  => false
        case Skip  => false
        case Cons(_, _) => false
        case Dispose(_)  => false
        case Seq(_, _) => false
        case If(_, _, _)  => false
        case While(_, _) => false
        case Atomic(_, _) => false
        // End of not used =============

        case Par(C1, C2) => (
                exists p1, p2, q1, q2, r, r1', r2', m, m1', m2',
                        G1, G2, I, G1', G2', I' ::
                G == Action.Or(G1, G2)
                && p == Assertion.Indep(p1, Assertion.Indep(p2, Assertion.Indep(m, r)))
                && q == Assertion.Indep(q1, Assertion.Indep(q2,
                    Assertion.Indep(Assertion.And(m1', m2'), Assertion.And(r1', r2'))))
                && ConcurrentlyDerivable(
                        Action.Indep(Action.Or(R, G2), G2'),
                        Action.Indep(G1, G1'),
                        Assertion.Indep(I, I'),
                        Assertion.Indep(p1, Assertion.Indep(m, r)),
                        C1,
                        Assertion.Indep(q1, Assertion.Indep(m1', r1'))
                )
                && ConcurrentlyDerivable(
                        Action.Indep(Action.Or(R, G1), G1'),
                        Action.Indep(G2, G2'),
                        Assertion.Indep(I, I'),
                        Assertion.Indep(p2, Assertion.Indep(m, r)),
                        C2,
                        Assertion.Indep(q2, Assertion.Indep(m2', r2'))
                )
                && Fence(I, R)
                && Fence(I, G1)
                && Fence(I, G2)
                && Fence(I', G1')
                && Fence(I, G2')
                && AssertionImpl(OrFromSeq([r, r1', r2']), I)
                && AssertionImpl(OrFromSeq([m, m1', m2']), I')
        )
        )
        // ENV-SHARE
        || (exists p', q', r ::
            p' == Assertion.Indep(p, r)
            && q' == Assertion.Indep(q, r)
            && SequentiallyDerivable(p', C, q')
            && Sta(r, Action.Indep(R, Id()))
            && Fence(I, R)
            && Fence(I, G)
            && AssertionImpl(r, Assertion.Indep(I, Bool(True)))
        )
        // FR-PRIVATE
        || (exists p', q', r ::
            p' == Assertion.Indep(p, r)
            && q' == Assertion.Indep(q, r)
            && ConcurrentlyDerivable(R, G, I, p', C, q')
        )
        // FR-SHARE
        || (exists p', q', r, R', R'', G', G'', I', I'' ::
            p' == Assertion.Indep(p, r)
            && q' == Assertion.Indep(q, r)
            && R == Action.Indep(R', R'')
            && G == Action.Indep(G', G'')
            && I == Assertion.Indep(I', I'')
            && ConcurrentlyDerivable(R', G', I', p', C, q')
            && Sta(r, R'')
            && Fence(I'', R'')
            && Fence(I'', G'')
            && AssertionImpl(r, I'')
        )
    }
}
