include "util.dfy"

module programming_language {
    import opened util

    // ==========================================================================
    // Programming language
    // ==========================================================================

    type VName = string

     // (Expr)   E ::= x | n | E + E | ...
    datatype Expr = Var(VName) | LVar(VName) | Int(int) | Add(Expr, Expr)
        | Substract(Expr, Expr)

    function ValExpr(expr: Expr, s: Store, i: LvMap): int
        requires DefinedExpr(expr, s, i)
    {
        match expr
        case Var(x) => s[x]
        case LVar(x) => i[x]
        case Int(n) => n
        case Add(e0, e1) => ValExpr(e0, s, i) + ValExpr(e1, s, i)
        case Substract(e0, e1) => ValExpr(e0, s, i) + ValExpr(e1, s, i)
    }

    function FvExpr(expr: Expr):set<VName> {
        match expr
        case Var(x) => {}
        case LVar(x) => {x}
        case Int(n) => {}
        case Add(e0, e1) => FvExpr(e0) + FvExpr(e1)
        case Substract(e0, e1) => FvExpr(e0) + FvExpr(e1)
    }

    predicate DefinedExpr(expr: Expr, s: Store, i: LvMap) {
        match expr
        case Var(x) => x in s
        case LVar(x) => x in i
        case Int(n) => true
        case Add(e0, e1) => DefinedExpr(e0, s, i) && DefinedExpr(e1, s, i)
        case Substract(e0, e1) => DefinedExpr(e0, s, i) && DefinedExpr(e1, s, i)
    }

    function ReplaceVar(E: Expr, o: VName, n: VName): Expr {
        match E
        case Var(x) => if x == o then LVar(n) else Var(x)
        case LVar(x) => E
        case Int(n) => E
        case Add(e0, e1) => Add(ReplaceVar(e0, o, n), ReplaceVar(e1, o, n))
        case Substract(e0, e1) => Substract(ReplaceVar(e0, o, n), ReplaceVar(e1, o, n))
    }

    // (BExpr)  B ::= true | false | E = E | !B | ...
    datatype BExpr = True | False
        | Not(BExpr) | And(BExpr, BExpr) | Or(BExpr, BExpr) | Xor(BExpr, BExpr)
        | Equal(Expr, Expr) | Greater(Expr, Expr) | GreaterEqual(Expr, Expr)
        | Less(Expr, Expr) | LessEqual(Expr, Expr)

    function ValBExpr(bexpr: BExpr, s: Store, i: LvMap):bool
        requires DefinedBExpr(bexpr, s, i)
    {

        match bexpr
        case True => true
        case False => false
        case Not(b) => !ValBExpr(b, s, i)
        case And(b0, b1) => ValBExpr(b0, s, i) && ValBExpr(b1, s, i)
        case Or(b0, b1) => ValBExpr(b0, s, i) || ValBExpr(b1, s, i)
        case Xor(b0, b1) => !(ValBExpr(b0, s, i) <==> ValBExpr(b1, s, i))
        case Equal(e0, e1) => ValExpr(e0, s, i) == ValExpr(e1, s, i)
        case Greater(e0, e1) => ValExpr(e0, s, i) > ValExpr(e1, s, i)
        case GreaterEqual(e0, e1) => ValExpr(e0, s, i) >= ValExpr(e1, s, i)
        case Less(e0, e1) => ValExpr(e0, s, i) < ValExpr(e1, s, i)
        case LessEqual(e0, e1) => ValExpr(e0, s, i) <= ValExpr(e1, s, i)
    }

    function FvBExpr(bexpr: BExpr):set<VName> {
        match bexpr
        case True => {}
        case False => {}
        case Not(b) => FvBExpr(b)
        case And(b0, b1) => FvBExpr(b0) + FvBExpr(b1)
        case Or(b0, b1) => FvBExpr(b0) + FvBExpr(b1)
        case Xor(b0, b1) => FvBExpr(b0) + FvBExpr(b1)
        case Equal(e0, e1) => FvExpr(e0) + FvExpr(e1)
        case Greater(e0, e1) => FvExpr(e0) + FvExpr(e1)
        case GreaterEqual(e0, e1) => FvExpr(e0) + FvExpr(e1)
        case Less(e0, e1) => FvExpr(e0) + FvExpr(e1)
        case LessEqual(e0, e1) => FvExpr(e0) + FvExpr(e1)
    }

    predicate DefinedBExpr(bexpr: BExpr, s: Store, i: LvMap) {
        match bexpr
        case True => true
        case False => true
        case Not(b) => DefinedBExpr(b, s, i)
        case And(b0, b1) => DefinedBExpr(b0, s, i) && DefinedBExpr(b1, s, i)
        case Or(b0, b1) => DefinedBExpr(b0, s, i) && DefinedBExpr(b1, s, i)
        case Xor(b0, b1) => DefinedBExpr(b0, s, i) && DefinedBExpr(b1, s, i)
        case Equal(b0, b1) => DefinedExpr(b0, s, i) && DefinedExpr(b1, s, i)
        case Greater(b0, b1) => DefinedExpr(b0, s, i) && DefinedExpr(b1, s, i)
        case GreaterEqual(b0, b1) => DefinedExpr(b0, s, i) && DefinedExpr(b1, s, i)
        case Less(b0, b1) => DefinedExpr(b0, s, i) && DefinedExpr(b1, s, i)
        case LessEqual(b0, b1) => DefinedExpr(b0, s, i) && DefinedExpr(b1, s, i)
    }

    // (Stmt)  C ::= x := E | x := [E] | [E] := E | ...
    datatype Stmt = Assign(AssignDst: VName, AssignVal: Expr)
        | Load(LoadDst: VName, LoadAddr: Expr)
        | Store(StoreAddr: Expr, StoreVal: Expr)
        | Skip
        | Cons(ConsDst: VName, ConsVals: seq<Expr>)
        | Dispose(DisposeAddr: Expr)
        | Seq(SeqC1: Stmt, SeqC2: Stmt)
        | If(IfCond: BExpr, IfC1: Stmt, IfC2: Stmt)
        | While(WhileCond: BExpr, WhileC: Stmt)
        | Atomic(AtomicCond: BExpr, AtomicC: Stmt)
        | Par(ParC1: Stmt, ParC2: Stmt)


    // ==========================================================================
    // Stores
    // ==========================================================================

    type Var = string

    // (Store)  s ∈ Var-->Int
    type Store = map<Var, int>
    // (LvMap)  i ∈ Var-->Int
    type LvMap = map<Var, int>
    // (Heap)   h ∈ Nat-->Int
    type Heap = map<nat, int>

    type Mem = (Store, LvMap, Heap)
}

