include "programming_language.dfy"
include "util.dfy"

module assertions {
    import opened programming_language
    import opened util

    // ================================================================
    // Syntax of the assertion language
    // ================================================================

    // (RelAssn)  P,Q,J ::= B | emp | E +-> E | E => E
    //     | ⌊⌊p⌋⌋ | P*Q | P ∧ Q | P ∨ Q | ...

    datatype RelAssn = Bool(BExpr) | Emp | Points(Expr, Expr) | PointsAbs(Expr, Expr)
        | ToRelAssn(FullAssn) | Indep(RelAssn, RelAssn) | And(RelAssn, RelAssn) | Or(RelAssn, RelAssn)
        | ForAll(VName, RelAssn)

    predicate Semantic_RelAssn(assrt: RelAssn, mem: AllMem)
    {
        var s0 := mem.0;
        var h0 := mem.1;
        var s1 := mem.2;
        var h1 := mem.3;
        match assrt
        case Bool(b) => s0 !! s1 && BExprVal(b, union(s0, s1))
        case Emp => |h0| == 0 && |h1| == 0
        case Points(e0, e1) => |h0| == 1
            && s0 !! s1
            && ExprVal(e0, union(s0, s1)) in h0
            && h0[ExprVal(e0, union(s0, s1))] == ExprVal(e1, union(s0, s1))
        case PointsAbs(e0, e1) => |h1| == 1
            && s0 !! s1
            && ExprVal(e0, union(s0, s1)) in h1
            && h1[ExprVal(e0, union(s0, s1))] == ExprVal(e1, union(s0, s1))
        case ToRelAssn(p) => exists u, w, C :: 0 <= u && 0 <= w && Semantic_FullAssn(p, mem, u, w, C)
        case Indep(P, Q) =>
            exists m, m' :: mem_unionable(m, m')
            && mem == mem_union(m, m')
            && Semantic_RelAssn(P, m)
            && Semantic_RelAssn(Q, m')
        case And(P, Q) => Semantic_RelAssn(P, mem) && Semantic_RelAssn(Q, mem)
        case Or(P, Q) => Semantic_RelAssn(P, mem) || Semantic_RelAssn(Q, mem)
        case ForAll(x, P) =>
            forall n: int :: mem_check_var(mem, x, n)
            ==>
            Semantic_RelAssn(P, mem)
    }

    function fv_RelAssn(assrt: RelAssn):set<VName> {
        match assrt
        case Bool(b) => fv_BExpr(b)
        case Emp => {}
        case Points(e0, e1) => fv_Expr(e0) + fv_Expr(e1)
        case PointsAbs(e0, e1) => fv_Expr(e0) + fv_Expr(e1)
        case ToRelAssn(p) => fv_FullAssn(p)
        case Indep(P, Q) => fv_RelAssn(P) + fv_RelAssn(Q)
        case And(P, Q) => fv_RelAssn(P) + fv_RelAssn(Q)
        case Or(P, Q) => fv_RelAssn(P) + fv_RelAssn(Q)
        case ForAll(x, P) => fv_RelAssn(P)
    }


    // (RelAct) R,G ::= P ⋉_k Q | [P] | 𝓓
    //     | ⌊G⌋₀ | G ∧ G | G ∨ G | ...

    datatype RelAct = Action(RelAssn, RelAssn) | Mantain(RelAssn) | '(DAct)
        | /* ?? |*/ And(RelAct, RelAct) | Or(RelAct, RelAct) | ForAll(VName, RelAct)

    predicate Semantic_RelAct(assrt: RelAct, mem0: AllMem, mem1: AllMem)
    {
        match assrt
        case Action(P, Q) => Semantic_RelAssn(P, mem0) && Semantic_RelAssn(Q, mem1)
        case Mantain(P) => mem0 == mem1 && Semantic_RelAssn(P, mem0)
        case '(D) => Semantic_DAct(D, mem0, mem1)
        //case  ?? =>  TODO
        case And(G, G') => Semantic_RelAct(G, mem0, mem1) && Semantic_RelAct(G', mem0, mem1)
        case Or(G, G') => Semantic_RelAct(G, mem0, mem1) && Semantic_RelAct(G', mem0, mem1)
        case ForAll(x, G) =>
            forall n: int ::
            mem_check_var(mem0, x, n) && mem_check_var(mem1, x, n)
            ==>
            Semantic_RelAct(G, mem0, mem1)
    }

    function Id():RelAct {
        Mantain(Bool(True))
    }

    function fv_RelAct(assrt: RelAct):set<VName> {
        match assrt
        case Action(P, Q) => fv_RelAssn(P) + fv_RelAssn(Q)
        case Mantain(P) => fv_RelAssn(P)
        case '(D) => fv_DAct(D)
        case And(G, G') => fv_RelAct(G) + fv_RelAct(G')
        case Or(G, G') => fv_RelAct(G) + fv_RelAct(G')
        case ForAll(x, G) => fv_RelAct(G)
    }


    // (DAct) ::= P ⤳ Q | ∀x.𝓓 | 𝓓 ∧ 𝓓
    datatype DAct = Action(RelAssn, RelAssn) | ForAll(VName, DAct) | And(DAct, DAct)

    predicate Semantic_DAct(assrt: DAct, mem0: AllMem, mem1: AllMem)
    {
        match assrt
        case Action(P, Q) => Semantic_RelAssn(P, mem0) ==> Semantic_RelAssn(Q, mem1)
        case ForAll(x, D) =>
            forall n: int ::
            mem_check_var(mem0, x, n) && mem_check_var(mem1, x, n)
            ==>
            Semantic_DAct(D, mem0, mem1)
        case And(D0, D1) => Semantic_DAct(D0, mem0, mem1) && Semantic_DAct(D1, mem0, mem1)
    }

    // Enabled(P ⤳ Q) = P
    // Enabled(∀x.𝓓) = ∀x.Enabled(𝓓)
    // Enabled(𝓓 ∧ 𝓓) = Enabled(𝓓) ∨ Enabled(𝓓)
    function Enabled(assrt: DAct):RelAssn {
        match assrt
        case Action(P, Q) => P
        case ForAll(x, D) => RelAssn.ForAll(x, Enabled(D))
        case And(D0, D1) => RelAssn.Or(Enabled(D0), Enabled(D1))
    }

    // <𝓓> = 𝓓 ∧ (Enabled(𝓓) ⋉ true)
    function AngleBracket(D: DAct):RelAct {
        RelAct.And(RelAct.'(D), RelAct.Action(Enabled(D), Bool(True)))
    }

    // [𝓓] = Enabled(𝓓) ⤳ Enabled(𝓓)
    function Bracket(D: DAct):DAct {
        DAct.Action(Enabled(D), Enabled(D))
    }

    // 𝓓' ≤ 𝓓 iff (Enabled(𝓓') ==> Enabled(𝓓)) ∧ (𝓓 ==> 𝓓')
    predicate LeqDAct(D: DAct, D': DAct) {
        (forall m :: Semantic_RelAssn(Enabled(D'), m) ==> Semantic_RelAssn(Enabled(D), m))
        && (forall m0, m1 :: Semantic_DAct(D, m0, m1) ==> Semantic_DAct(D', m0, m1))
    }

    function fv_DAct(assrt: DAct):set<VName> {
        match assrt
        case Action(P, Q) => fv_RelAssn(P) + fv_RelAssn(Q)
        case ForAll(x, D) => fv_DAct(D)
        case And(D0, D1) => fv_DAct(D0) + fv_DAct(D1)
    }


    // (FullAssn) p,q ::= P | arem(C) | ◊(E) | ⧫(E_k, ..., E_1)
    //     |  ⌊p⌋_a | ⌊p⌋_◊ | p * q | p ∧ q | p ∨ q | ...
    datatype FullAssn = '(RelAssn) | Arem(Stmt) | White(Expr) | Black(seq<Expr>)
        | IgnoreStmt(FullAssn) | IgnoreWhite(FullAssn) | Indep(FullAssn, FullAssn)
        | And(FullAssn, FullAssn) | Or(FullAssn, FullAssn) | ForAll(VName, FullAssn)

    predicate Semantic_FullAssn(assrt: FullAssn, mem: AllMem, u: int, w: int, C: Stmt)
        requires u >= 0 && w >= 0
        decreases assrt
    {
        match assrt
        case '(P) => Semantic_RelAssn(P, mem)
        case Arem(C') => C == C
        case White(e) => ExprVal(e, mem.0) <= w
        case Black(se) => forall e :: e in se ==> ExprVal(e, mem.0) <= u
        case IgnoreStmt(p) => exists w' :: w' >= 0 && Semantic_FullAssn(p, mem, u, w', C)
        case IgnoreWhite(p) => exists C' :: Semantic_FullAssn(p, mem, u, w, C')
        case Indep(p, q) =>
            exists C0, C1, m0, m1, u0, u1, w0, w1 ::
            ((C0 == C && C1 == Stmt.Skip) || (C1 == C && C0 == Stmt.Skip))
            && u0 >= 0 && u1 >= 0 && u0 + u1 == u
            && w0 >= 0 && w1 >= 0 && w0 + w1 == w
            && mem_unionable(m0, m1) && mem_union(m0, m1) == mem
            && Semantic_FullAssn(p, m0, u0, w0, C0)
            && Semantic_FullAssn(q, m1, u1, w1, C1)
        case And(p, q) => Semantic_FullAssn(p, mem, u, w, C) && Semantic_FullAssn(q, mem, u, w, C)
        case Or(p, q) => Semantic_FullAssn(p, mem, u, w, C) || Semantic_FullAssn(q, mem, u, w, C)
        case ForAll(x, p) =>
            forall n: int :: mem_check_var(mem, x, n)
            ==>
            Semantic_FullAssn(p, mem, u, w, C)
    }


    function fv_FullAssn(assrt: FullAssn):set<VName> {
        match assrt
        case '(P) => fv_RelAssn(P)
        case Arem(C') => {}
        case White(e) => fv_Expr(e)
        case Black(se) => fv_seq_Expr(se)
        case IgnoreStmt(p) => fv_FullAssn(p)
        case IgnoreWhite(p) => fv_FullAssn(p)
        case Indep(p, q) => fv_FullAssn(p) + fv_FullAssn(q)
        case And(p, q) => fv_FullAssn(p) + fv_FullAssn(q)
        case Or(p, q) => fv_FullAssn(p) + fv_FullAssn(q)
        case ForAll(x, p) => fv_FullAssn(p)
    }

    function fv_seq_Expr(se: seq<Expr>):set<VName> {
        if |se| == 0 then {}
        else fv_Expr(se[0]) + fv_seq_Expr(se[1..])
    }


    // ================================================================
    // Additional state to define assertion semantics
    // ================================================================

    // (Real Store, Real Heap, Symbolic Store, Symbolic Heap)
    type AllMem = (Store, Heap, Store, Heap)

    function MemToAllMem(m0: Mem, m1: Mem): AllMem {
        (m0.0, m0.1, m1.0, m1.1)
    }

    function MemUnion(m: Mem, m': Mem): Mem
        requires MemUnionable(m, m')
    {
        var s := m.0;
        var h := m.1;
        var s' := m'.0;
        var h' := m'.1;
        (s, union(h, h'))
    }

    function mem_union(mem: AllMem, mem': AllMem):AllMem
        requires mem_unionable(mem, mem')
    {
        var s0 := mem.0;
        var h0 := mem.1;
        var s1 := mem.2;
        var h1 := mem.3;

        var s0' := mem'.0;
        var h0' := mem'.1;
        var s1' := mem'.2;
        var h1' := mem'.3;

        (s0, union(h0, h0'), s1, union(h1, h1'))
    }

    predicate MemUnionable(m: Mem, m': Mem) {
        // Stores are equal between them
        m.0 == m'.0
        // Heaps and independent between them
        && m.1 !! m'.1
    }

    predicate mem_unionable(mem: AllMem, mem': AllMem) {
        // Real stores and symbolic stores are equal between them
        mem.0 == mem'.0
        && mem.2 == mem'.2
        // Real heaps and symbolic heaps are independent between them
        && mem.1 !! mem'.1
        && mem.3 !! mem'.3
    }

    predicate mem_check_var(mem: AllMem, x: VName, v: int) {
        x in mem.0 && mem.0[x] == v
    }
}

