module programming_language {

    // ==========================================================================
    // Programming language
    // ==========================================================================

    type VName = string

    // (MName)  f  ∈ String
    type MName = string

    // (Expr)   E,𝔼  ::= x | n | E + E | ...
    datatype Expr = Var(VName) | Nat(int) | Add(Expr, Expr) | Substract(Expr, Expr)

    function ExprVal(expr: Expr, s: Store): int
    {
        match expr
        case Var(x) => if x in s then s[x] else 0  // TODO: else error checking
        case Nat(n) => n
        case Add(e0, e1) => ExprVal(e0, s) + ExprVal(e1, s)
        case Substract(e0, e1) => ExprVal(e0, s) + ExprVal(e1, s)
    }

    function fv_Expr(expr: Expr):set<VName> {
        match expr
        case Var(x) => {x}
        case Nat(n) => {}
        case Add(e0, e1) => fv_Expr(e0) + fv_Expr(e1)
        case Substract(e0, e1) => fv_Expr(e0) + fv_Expr(e1)
    }

    predicate defined(e: Expr, s: Store) {
        match e
        case Var(x) => x in s
        case Nat(n) => true
        case Add(e0, e1) => defined(e0, s) && defined(e1, s)
        case Substract(e0, e1) => defined(e0, s) && defined(e1, s)
    }

    // (BExpr)  B,𝔹  ::= true | false | E = E | !B | ...
    datatype BExpr = True | False
        | Not(BExpr) | And(BExpr, BExpr) | Or(BExpr, BExpr) | Xor(BExpr, BExpr)
        | Equal(Expr, Expr) | Greater(Expr, Expr) | GreaterEqual(Expr, Expr)
        | Less(Expr, Expr) | LessEqual(Expr, Expr)

    function BExprVal(bexpr: BExpr, s: Store):bool
    {
        match bexpr
        case True => true
        case False => false
        case Not(b) => !BExprVal(b, s)
        case And(b0, b1) => BExprVal(b0, s) && BExprVal(b1, s)
        case Or(b0, b1) => BExprVal(b0, s) || BExprVal(b1, s)
        case Xor(b0, b1) => !(BExprVal(b0, s) <==> BExprVal(b1, s))
        case Equal(e0, e1) => defined(e0, s) && defined(e1, s) && ExprVal(e0, s) == ExprVal(e1, s)
        case Greater(e0, e1) => defined(e0, s) && defined(e1, s) && ExprVal(e0, s) > ExprVal(e1, s)
        case GreaterEqual(e0, e1) => defined(e0, s) && defined(e1, s) && ExprVal(e0, s) >= ExprVal(e1, s)
        case Less(e0, e1) => defined(e0, s) && defined(e1, s) && ExprVal(e0, s) < ExprVal(e1, s)
        case LessEqual(e0, e1) => defined(e0, s) && defined(e1, s) && ExprVal(e0, s) <= ExprVal(e1, s)
    }

    function fv_BExpr(bexpr: BExpr):set<VName> {
        match bexpr
        case True => {}
        case False => {}
        case Not(b) => fv_BExpr(b)
        case And(b0, b1) => fv_BExpr(b0) + fv_BExpr(b1)
        case Or(b0, b1) => fv_BExpr(b0) + fv_BExpr(b1)
        case Xor(b0, b1) => fv_BExpr(b0) + fv_BExpr(b1)
        case Equal(e0, e1) => fv_Expr(e0) + fv_Expr(e1)
        case Greater(e0, e1) => fv_Expr(e0) + fv_Expr(e1)
        case GreaterEqual(e0, e1) => fv_Expr(e0) + fv_Expr(e1)
        case Less(e0, e1) => fv_Expr(e0) + fv_Expr(e1)
        case LessEqual(e0, e1) => fv_Expr(e0) + fv_Expr(e1)
    }

    // (Instr)  c,𝕔  ::= x := E | x := [E] | [E] := E | print(E)
    datatype Instr = Assign(VName, Expr) | MemRead(VName, Expr)
        | MemWrite(Expr, Expr) | Print(Expr) | Cons(VName, seq<Expr>)
        | Dispose(Expr)


    predicate InstrStep(instr: Instr, m0: Mem, m1: Mem)
    {
        var s0 := m0.0;
        var h0 := m0.1;
        var s1 := m1.0;
        var h1 := m1.1;

        match instr
        case Assign(x, e) =>
            s1 == s0[x := ExprVal(e, s0)]
        case MemRead(x, e) =>
            var addr := ExprVal(e, s0);
            addr in h0 && s1 == s0[x := h0[addr]]
        case MemWrite(e0, e1) =>
            h1 == h0[ExprVal(e0, s0) := ExprVal(e1, s0)]
        case Print(e) =>
            // TODO: sidefect
            // print ExprVal(e) ?
            m1 == m0
        case Cons(x, se) =>
            exists addr :: 0 <= addr && ConsHelper(addr, se, s0, h0, h1)
        case Dispose(e) =>
            var addr := ExprVal(e, s0);
            addr in h0 && addr !in h1
    }

    predicate ConsHelper(addr: int, se: seq<Expr>, s: Store, h0: Heap, h1: Heap)
        decreases |se|
    {
        if  |se| == 0 then
            true
        else
            addr !in h0
                && addr in h1
                && h1[addr] == ExprVal(se[0], s)
                && ConsHelper(addr+1, se[1..], s, h0, h1)
    }


    // (Stmt)   C,ℂ  ::= skip | c | x := f(E) | return E | <C>
    datatype Stmt = Skip | Instr(Instr) | Call(VName, MName, Expr) | Return(Expr)
        | Atomic(Stmt) | End | Compose(Stmt, Stmt) | Cond(BExpr, Stmt, Stmt)
        | While(BExpr, Stmt)


    inductive predicate StmtStep (stmt: Stmt, m0: Mem, m1: Mem)
    {
        var s0 := m0.0;
        var h0 := m0.1;
        var s1 := m1.0;
        var h1 := m1.1;

        match stmt
        case Skip =>
            m0 == m1
        case Instr(instr) =>
            InstrStep(instr, m0, m1)
        case Call(x, f, e) =>
            // TODO
            m0 == m1
        case Return(e) =>
            // TODO
            m0 == m1
        case Atomic(stmt) =>
            // TODO
            StmtStep(stmt, m0, m1)
        case End =>
            // TODO
            true
        case Compose(stmt0, stmt1) =>
            exists m' ::
            StmtStep(stmt0, m0, m') &&
            StmtStep(stmt1, m', m1)
        case Cond(b, thn, els) =>
            StmtStep(if BExprVal(b, s0) then thn else els, m0, m1)
        case While(b, body) =>
            (!BExprVal(b, s0) && m0 == m1) ||
            (BExprVal(b, s0) && exists m' ::
                StmtStep(body, m0, m') &&
                StmtStep(While(b, body), m', m1))
    }

    // (Prog)   W,𝕎  ::= skip | let Π in C ||...||C   (Unused)
    //datatype Program = Skip | Let(ODecl, seq<Stmt>)

    // (ODecl)  Π,Γ  ::= {f1 -> (x1,C1), ..., fn -> (xn,Cn)}
    type ODecl = map<MName, (VName, Stmt)>


    // ==========================================================================
    // States and traces
    // ==========================================================================

    type Var = string

    // (ThrdID)  t ∈ Nat
    type ThrdID = nat
    // (ExecCtxt)  E ::= [] | E;C
    type ExecCtxt = seq<Stmt>

    // (Store)  s,𝕤 ∈ Var-->Int
    type Store = map<Var, int>
    // (Heap)  h,𝕙 ∈ Nat-->Int
    type Heap = map<nat, int>
    // (Mem)  σ,Σ ::= (s, h)
    type Mem = (Store, Heap)
    // (CallStk)  k,𝕜 := (sl, x, C) | ◦
    datatype CallStk = Stack(Store, Var, Stmt) | Nil

    // (ThrdPool)  K,K ::= (t1 -> k1, ..., tn -> kn)
    type ThrdPool = seq<(ThrdID, CallStk)>
    // (PState)  S,S ::= (σ_c, σ_o, K)
    type PState = (Mem, Mem, ThrdPool)


    // (Evt)  e ::= (t, f, n) | (t, ret, n) | (t, obj) | (t, clt)
    //     | (t, out, n) | (t, term) | (spawn, n)
    //     | (t, obj, abort) | (t, clt, abort)
    datatype Evt =
          '(ThrdID, MName, int) | Ret(ThrdID, int) | Obj(ThrdID) | Clt(ThrdID)
        | Out(ThrdID, int) | Term(ThrdID) | Spawn(int)
        | ObjAbort(ThrdID) | CltAbort(ThrdID)

    // (ETrace) T, T ::= 𝜖 | e::T (co-inductive)
    codatatype ETrace = Nil | Trace(Evt, ETrace)
}

